/*
 * Copyright C.L.U.T. - Torino - Ottobre 2017
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 * chapter 01 - exercise 21 - solution (2-1)
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "item.h"
#include "symbolTable.h"
#include "graph.h"

struct graph_s {
  int V;
  int E;
  int **mAdj;
};

// Prototypes
static void graph_insert_edge(graph_t, int, int, int);
static void graph_max_path_find_r(graph_t, int, int *, int, int, int *, int *,
  int *, st_t);
static void powerset(graph_t, st_t, int, int *, int *, int *, int);
static int check(graph_t, st_t, int, int *, int *);
static int constraint_check(graph_t, st_t, int *);
static void connectivity_check(graph_t, int, int *, int *);

/*
 *  Read graph from file
 */
graph_t graph_read(char *name, st_t *st) {
  FILE *fp;
  int V = -1, wt, id1, id2;
  char src[MAXL], dst[MAXL], c1[10], c2[10];
  graph_t graph = NULL;

  fp = fopen(name, "r");
  if (fp==NULL) {
    fprintf (stderr, "File open error.\n");
    exit(EXIT_FAILURE);
  }

  *st = st_init(1);
  if (*st == NULL)
    return NULL;

  while(fscanf(fp, "%s %s %d %s %s", src, c1, &wt, dst, c2) == 5) {
    id1 = st_search(*st, src);
    id2 = st_search(*st, dst);
    if (id1 == -1) {
      st_insert(*st, src, COLOR(c1));
    } else {
      if (st_get_color(*st, id1) != COLOR(c1)) {
        fprintf (stderr, "Wrong color for node %s.\n", src);
        exit(EXIT_FAILURE);
      }
    }
    if (id2 == -1) {
      st_insert(*st, dst, COLOR(c2));
    } else {
      if (st_get_color(*st, id2) != COLOR(c2)) {
        fprintf (stderr, "Wrong color for node %s.\n", src);
        exit(EXIT_FAILURE);
      }
    }
  }

  V = st_get_num(*st);
  graph = graph_init(V);

  if (graph == NULL) {
    return NULL;
  }

  fclose (fp);
  fp = fopen(name, "r");

  while(fscanf(fp, "%s %s %d %s %s", src, c1, &wt, dst, c2) == 5) {
    id1 = st_search(*st, src);
    id2 = st_search(*st, dst);
    graph_insert_edge(graph, id1, id2, wt);
    graph_insert_edge(graph, id2, id1, wt);
    graph->E++;
  }

  fclose (fp);

  return graph;
}

graph_t graph_init(int V) {
  int v;
  graph_t graph;

  graph = malloc(sizeof *graph);
  if (graph == NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }
  graph->V = V;
  graph->E = 0;
  graph->mAdj = calloc(graph->V, sizeof(int*));
  if (graph->mAdj==NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }
  for (v = 0; v < graph->V; v++) {
    graph->mAdj[v] = calloc(graph->V, sizeof(int));
    if (graph->mAdj[v]==NULL) {
      fprintf (stderr, "Memory allocation error.\n");
      exit(EXIT_FAILURE);
    }
  }

  return graph;
}

static void graph_insert_edge(graph_t g, int x, int y, int wt) {
  g->mAdj[x][y] = g->mAdj[y][x] = wt;
}

int graph_get_nv(graph_t graph) {
  if(graph == NULL) {
    return -1;
  }

  return graph->V;
}

/*
 *  Find maximum weigth path
 */
void graph_max_path_find (graph_t g, st_t st) {
  int i, j, max = -1, best_max = -1;
  int *visited = calloc(g->V, sizeof(int));
  int *currpath = calloc(g->V, sizeof(int));
  int *bestpath = calloc(g->V, sizeof(int));
  int *best = calloc(g->V, sizeof(int));

  if (visited==NULL || currpath==NULL || bestpath==NULL || best==NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }

  for(i=0;i<g->V;i++) {
    max = -1;
    for(j=0;j<g->V;j++) {
      currpath[j] = bestpath[j] = -1;
      visited[j] = 0;
    }
    graph_max_path_find_r(g, i, &max, 0, 0, visited, currpath, bestpath, st);
    if (best_max < max) {
      best_max = max;
      memcpy(best, bestpath, g->V*sizeof(int));
    }
  }

  fprintf (stdout, "Path: ");
  i = 0;
  while(best[i] != -1) {
    fprintf (stdout, "%s(%s) ",
      st_get_name(st, best[i]), st_get_color(st, best[i]) ? "N" : "R");
    i++;
  }
  fprintf (stdout, "; path length = %d\n", best_max);

  free(visited);
  free(currpath);
  free(bestpath);
  free(best);
}

/*
 *  Recursive function to find maximum weigth path
 */
static void graph_max_path_find_r(
  graph_t g, int v, int *max, int currval, int pos,
  int *visited, int *path, int *bestpath, st_t st
) {
  int w;

  visited[v] = 1;
  path[pos] = v;

  if (1) {
      path[pos] = v;
      if (currval > *max) {
          *max = currval;
          memcpy(bestpath, path, g->V*sizeof(int));
      }
  }

  for (w=0; w<g->V; w++) {
    if (g->mAdj[v][w] && !visited[w]) {
      if (st_get_color(st, v)==RED && st_get_color(st,w)==RED) {
        continue;
      }
      graph_max_path_find_r (g, w, max, currval+g->mAdj[v][w],
                             pos+1, visited, path, bestpath, st);
    }
  }

  visited[v] = 0;

  return;
}

/*
 *  Find connected sub-graph
 */
void  graph_subset_find(graph_t g, st_t st) {
  int i, best = 0;
  int *sub_graph = calloc(g->V, sizeof(int));
  int *sub_graph_best = calloc(g->V, sizeof(int));

  if (sub_graph==NULL || sub_graph_best==NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }

  powerset(g, st, 0, sub_graph, sub_graph_best, &best, 0);
  fprintf (stdout, "Vertex in sub-graph: ");
  for(i=0;i<g->V;i++) {
    if (sub_graph_best[i]) {
      fprintf (stdout, "(%s) ", st_get_name(st, i));
    }
  }
  fprintf (stdout, "; value = %d\n", best);

  free(sub_graph);
  free(sub_graph_best);

  return;
}

/*
 *  Generate powerset
 */
static void powerset(
  graph_t g, st_t st, int pos, int *sub_graph, int *sub_graph_best,
  int *best, int last
) {
  int curr = 0;
  if (last == 1 && check(g, st, pos, sub_graph, &curr)) {
    if (*best < curr) {
      *best = curr;
      memcpy(sub_graph_best, sub_graph, g->V*sizeof(int));
    }
  }
  if (pos>=g->V) {
    return;
  }

  sub_graph[pos] = 1;
  powerset(g, st, pos+1, sub_graph, sub_graph_best, best, 1);
  sub_graph[pos] = 0;
  powerset(g, st, pos+1, sub_graph, sub_graph_best, best, 0);

  return;
}

/*
 *  Check constraints on sub-graph
 */
static int check(graph_t g, st_t st, int pos, int *sub_graph, int *curr) {
  int ok = constraint_check(g, st, sub_graph);

  if (ok) {
    int i, j;
    for(i=0;i<g->V;i++) {
      if (sub_graph[i] == 0) continue;
      for(j=i+1;j<g->V;j++) {
        if (sub_graph[j] == 0 || !g->mAdj[i][j]) continue;
        (*curr) += g->mAdj[i][j];
      }
    }
    return 1;
  } else {
    return 0;
  }
}

static int constraint_check(graph_t g, st_t st, int *sub_graph) {
  int i, nR = 0, nN = 0, start = -1, ok = 1;
  int *visited;

  // color check
  for(i=0;i<g->V;i++) {
    if (sub_graph[i] == 1) {
      // start searh from mfirst vertex found
      if (start == -1) start = i;
      if (st_get_color(st, i) == RED) nR++;
      else { nN++; }
    }
  }
  if (start == -1 || abs(nN-nR) > 2) return 0;

  visited = calloc(g->V, sizeof(int));
  if (visited==NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }
  connectivity_check(g, start, sub_graph, visited);
  for(i=0;i<g->V;i++) {
    if (sub_graph[i] == 1 && !visited[i]) {
      ok = 0; break;
    }
  }
  free(visited);

  return ok;
}

/*
 *  Check connectivity with DFS
 */
static void connectivity_check(graph_t g, int start, int *sub_graph, int *visited) {
  visited[start] = 1;
  int i;
  for(i=0;i<g->V;i++) {
    if (g->mAdj[start][i] && sub_graph[i] && !visited[i]) {
      connectivity_check(g, i, sub_graph, visited);
    }
  }
}




/*
 * Copyright C.L.U.T. - Torino - Ottobre 2017
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 */
