/*
 * Copyright C.L.U.T. - Torino - Ottobre 2017
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 * chapter 01 - exercise 27 - solution (2-1)
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "graph.h"
#include "symbolTable.h"
#include "list.h"

typedef struct clique_s {
  list_t cl;
  int nNodi;
} clique_t;

/* prototypes */
static graph_t file_read(char *, st_t *);
static void clique_find(graph_t, st_t);
static void clique_find_r(graph_t, int *, int, int, st_t, clique_t *, int *, int *);
static void clique_write(clique_t *, int, st_t);
static int check(graph_t, int *);

int main(int argc, char **argv) {
  graph_t g = NULL;
  st_t ts = NULL;

  if (argc != 2) {
    fprintf (stderr, "Run as: %s <file_in>\n", argv[0]);
    exit(EXIT_FAILURE);
  }

  g = file_read(argv[1], &ts);
  if (g == NULL || ts == NULL) {
    fprintf (stderr, "Read file error.\n");
    exit(EXIT_FAILURE);
  }

  clique_find(g, ts);

  return 0;
}

static graph_t file_read(char *name, st_t *ts) {
  FILE *fp;
  graph_t g;
  int id1, id2;
  char c[MAXL], r[MAXL];

  fp = fopen(name, "r");
  if (fp == NULL) {
    fprintf (stderr, "File open error.\n");
    exit(EXIT_FAILURE);
  }

  *ts = st_init(1);
  if (*ts == NULL) {
    return NULL;
  }

  while(fscanf(fp, "%s %s", c, r) == 2) {
    id1 = st_find_or_add(*ts, c);
    id2 = st_find_or_add(*ts, r);
    if (id1 < 0 || id2 < 0) {
      fprintf (stderr, "Edge (%s, %s) does not exist. Ignored.\n", c, r);
    }
  }

  fclose(fp);
  fp = fopen(name, "r");

  g = graph_init(st_get_size(*ts));
  if (g == NULL) {
    return NULL;
  }

  while(fscanf(fp, "%s %s", c, r) == 2) {
    id1 = st_find(*ts, c);
    id2 = st_find(*ts, r);
    if (id1 == id2 || id1 < 0 || id2 < 0) {
      continue;
    }
    graph_insert_edge(g, id1, id2);
  }

  fclose(fp);

  return g;
}

/* Print-out a clique */
static void clique_write(clique_t *cliques, int nCliques, st_t ts) {
  int i;

  for (i=0; i<nCliques; i++) {
    fprintf (stdout, "Clique #%d:\n", i);
    list_write_name(cliques[i].cl, ts);
  }

  return;
}

/* Find cliques */
static void clique_find (graph_t g, st_t ts) {
  int *sol, nCliques = 0, largestIndex = -1;
  clique_t *cliques;

  sol = calloc(graph_get_nv(g), sizeof(int));
  cliques = calloc(graph_get_nv(g), sizeof(clique_t));
  if (sol==NULL || cliques==NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }

  clique_find_r(g, sol, 0, 0, ts, cliques, &nCliques, &largestIndex);

  fprintf (stdout, "Largests Maximal Clique: index=%d size=%d\n",
    largestIndex, cliques[largestIndex].nNodi);

  clique_write(cliques, nCliques, ts);

  return;
}

/* Find cliques; powerset adopting arrangements with repetitions */
static void clique_find_r(
  graph_t g, int *sol, int pos, int currNodi, st_t ts,
  clique_t *cliques, int *nCliques, int *largestIndex
) {
  int i, v = graph_get_nv(g);

  if (pos >= v) {
    if(check(g, sol)) {
      list_t newClique = list_init();
      for (i=0; i<v; i++)
        if (sol[i])
          list_insert_in_order(newClique, i);
      cliques[*nCliques].cl = newClique;
      cliques[*nCliques].nNodi = currNodi;

      if (*largestIndex == -1 || currNodi > cliques[*largestIndex].nNodi) {
        *largestIndex = *nCliques;
      }

      *nCliques = *nCliques +1;
    }
    return;
  }

  sol[pos] = 0;
  clique_find_r(g, sol, pos+1, currNodi, ts, cliques, nCliques, largestIndex);
  sol[pos] = 1;
  clique_find_r(g, sol, pos+1, currNodi+1, ts, cliques,
    nCliques, largestIndex);
}

/* Check sub-graph connectivity */
static int check(graph_t g, int *sol) {
  int i, j, v = graph_get_nv(g), c = 1;

  for (i=0; i<v-1; i++) {
    if (!sol[i])
      continue;
    for (j=i+1; j<v; j++) {
      if (!sol[j])
        continue;
      if (!graph_check_adjancency(g, i, j))
        return 0;
    }
  }

  for (i=0; i<v; i++) {
    if (sol[i])
      continue;
    c = 1;
    for (j=0; j<v; j++) {
      if (sol[j] && !graph_check_adjancency(g, i, j)) {
        c = 0;
        break;
      }
    }
    if (c == 1)
      return 0;
  }

  return 1;
}

/*
 * Copyright C.L.U.T. - Torino - Ottobre 2017
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 */
